from abc import ABC, abstractmethod
import base64
from datetime import datetime
from io import BytesIO
from pathlib import Path, PosixPath

import numpy as np
import pandas as pd


class ExternalTypeSerializer(ABC):
    class_: type
    class_name: str

    @classmethod
    @abstractmethod
    def to_json(cls, value: tuple) -> dict:
        pass

    @classmethod
    @abstractmethod
    def from_json(cls, value: tuple):
        pass


class TypeSerde(ExternalTypeSerializer):
    class_ = type
    class_name = 'type'

    mapping = {
        'bool': bool,
        'int': int,
        'float': float,
        'str': str,
        'float32': np.float32,
        'ValueError': ValueError
    }

    @classmethod
    def to_json(cls, value: type) -> dict:
        return {'value': value.__name__}

    @classmethod
    def from_json(cls, value: str) -> type:
        return cls.mapping[value]


class TupleSerde(ExternalTypeSerializer):
    class_ = tuple
    class_name = 'tuple'

    @classmethod
    def to_json(cls, value: tuple) -> dict:
        return {'as_list': list(value)}

    @classmethod
    def from_json(cls, as_list: list) -> tuple:
        return tuple(as_list)


class SliceSerde(ExternalTypeSerializer):
    class_ = slice
    class_name = 'slice'

    @classmethod
    def to_json(cls, value: slice) -> dict:
        return {'as_list': [value.start, value.stop, value.step]}

    @classmethod
    def from_json(cls, as_list: list) -> slice:
        return slice(*as_list)


class DateTimeSerde(ExternalTypeSerializer):
    class_ = datetime
    class_name = 'datetime'

    @classmethod
    def to_json(cls, value: datetime) -> dict:
        return {'isoformat': value.isoformat()}

    @classmethod
    def from_json(cls, isoformat: str) -> datetime:
        return datetime.fromisoformat(isoformat)


class PathSerde(ExternalTypeSerializer):
    class_ = PosixPath
    class_name = 'path'

    @classmethod
    def to_json(cls, value: Path) -> dict:
        return {'path_as_str': str(value)}

    @classmethod
    def from_json(cls, path_as_str: str) -> Path:
        return Path(path_as_str)


class NumpyScalarSerde(ExternalTypeSerializer):
    @classmethod
    def to_json(cls, value: np.number) -> dict:
        return {'value': value.item(), 'class_name': type(value).__name__}

    @classmethod
    def from_json(cls, value: str, class_name: str) -> np.number:
        return cls.class_(value)


class NumpyBoolSerde(NumpyScalarSerde):
    class_ = np.bool_
    class_name = 'np.bool_'


class NumpyU8Serde(NumpyScalarSerde):
    class_ = np.uint8
    class_name = 'np.uint8'


class NumpyU16Serde(NumpyScalarSerde):
    class_ = np.uint16
    class_name = 'np.uint16'


class NumpyU32Serde(NumpyScalarSerde):
    class_ = np.uint32
    class_name = 'np.uint32'


class NumpyU64Serde(NumpyScalarSerde):
    class_ = np.uint64
    class_name = 'np.uint64'


class NumpyI8Serde(NumpyScalarSerde):
    class_ = np.int8
    class_name = 'np.int8'


class NumpyI16Serde(NumpyScalarSerde):
    class_ = np.int16
    class_name = 'np.int16'


class NumpyI32Serde(NumpyScalarSerde):
    class_ = np.int32
    class_name = 'np.int32'


class NumpyI64Serde(NumpyScalarSerde):
    class_ = np.int64
    class_name = 'np.int64'


class NumpyF32Serde(NumpyScalarSerde):
    class_ = np.float32
    class_name = 'np.float32'

    @classmethod
    def to_json(cls, value: np.float32) -> dict:
        bytes_io = BytesIO()
        # noinspection PyTypeChecker
        np.save(bytes_io, value)
        o_bytes = bytes_io.getvalue()
        o_str = base64.b64encode(o_bytes).decode()
        return {'float_as_array': o_str}

    @classmethod
    def from_json(cls, float_as_array: str) -> np.float32:
        bytes_io = BytesIO(base64.b64decode(float_as_array.encode()))
        return np.load(bytes_io).flatten()[0]


class NumpyF64Serde(NumpyScalarSerde):
    class_ = np.float64
    class_name = 'np.float64'


class PandasMissingSerde(ExternalTypeSerializer):
    @classmethod
    def to_json(cls, value: np.number) -> dict:
        return {}

    @classmethod
    def from_json(cls) -> np.number:
        return cls.class_()


class PandasNASerde(PandasMissingSerde):
    class_ = pd._libs.missing.NAType
    class_name = 'pd.NA'


class PandasNaTSerde(PandasMissingSerde):
    class_ = pd._libs.tslibs.nattype.NaTType
    class_name = 'pd.NaT'


all_external_types = [
    TupleSerde, SliceSerde, DateTimeSerde, PathSerde, TypeSerde,
    NumpyBoolSerde, NumpyU8Serde, NumpyU16Serde, NumpyU32Serde, NumpyU64Serde,
    NumpyI8Serde, NumpyI16Serde, NumpyI32Serde, NumpyI64Serde, NumpyF32Serde, NumpyF64Serde,
    PandasNASerde, PandasNaTSerde
]
